"""
Tex Data Structures
---

Includes the data structures that users will interface with, in addition to
internally used data structures.
"""
import itertools

__all__ = ['TexNode', 'TexCmd', 'TexEnv', 'Arg', 'OArg', 'RArg', 'TexArgs']

#############
# Interface #
#############

class TexNode(object):
    """Main abstraction for Tex source, a tree node representing both Tex
    environments and Tex commands.
    """

    def __init__(self, expr):
        """Creates TexNode object

        :param (TexCmd, TexEnv) expr: a LaTeX expression, either a singleton
            command or an environment containing other commands
        """
        assert isinstance(expr, (TexCmd, TexEnv)), 'Created from TexExpr'
        super().__init__()
        self.expr = expr

    @property
    def name(self):
        return self.expr.name

    @property
    def args(self):
        return self.expr.args

    @property
    def parent(self):
        return self.expr.parent

    @property
    def string(self):
        """Returns 'string' content, which is valid if and only if (1) the
        expression is a TexCmd and (2) the command has only one argument.
        """
        if isinstance(self.expr, TexCmd) and len(self.expr.args) == 1:
            return str(self.expr.args[0])

    @property
    def tokens(self):
        """Returns generator of all tokens, for this Tex element"""
        return self.expr.tokens

    @property
    def contents(self):
        """Returns a generator of all contents, for this TeX element"""
        for child in self.expr.contents:
            if isinstance(child, (TexEnv, TexCmd)):
                yield TexNode(child)
            else:
                yield child

    @property
    def children(self):
        """Returns all immediate children of this TeX element"""
        for child in self.expr.children:
            child.parent = self
            yield TexNode(child)

    @property
    def descendants(self):
        """Returns all descendants for this TeX element."""
        return self.__descendants()

    def __descendants(self):
        """Implementation for descendants, hacky workaround for __getattr__
        issues.
        """
        return itertools.chain(self.contents,
            *[c.descendants for c in self.children])

    def find_all(self, name=None, **attrs):
        """Return all descendant nodes matching criteria, naively."""
        for descendant in self.__descendants():
            if hasattr(descendant, '__match__') and \
                descendant.__match__(name, attrs):
                yield descendant

    def find(self, name=None, **attrs):
        """Return first descendant node matching criteria"""
        try:
            return next(self.find_all(name, **attrs))
        except StopIteration:
            return None

    def count(self, name=None, **attrs):
        """Return number of descendants matching criteria"""
        return len(list(self.find_all(name, **attrs)))

    def __match__(self, name=None, attrs={}):
        r"""Check if given attributes match current object

        >>> from TexSoup import TexSoup
        >>> soup = TexSoup(r'\ref{hello}\ref{hello}\ref{hello}\ref{nono}')
        >>> soup.count(r'\ref{hello}')
        3
        """
        if '{' in name or '[' in name:
            return str(self) == name
        attrs['name'] = name
        for k, v in attrs.items():
            if getattr(self, k) != v:
                return False
        return True

    def __str__(self):
        """Stringified command"""
        return str(self.expr)

    def __repr__(self):
        """Interpreter representation"""
        return str(self)

    def __getattr__(self, attr, default=None):
        """Convert all invalid attributes into basic find operation."""
        return self.find(attr) or default

###############
# Expressions #
###############

class TexExpr(object):
    """General TeX expression abstract"""

    def __init__(self, name, contents=(), args=()):
        self.name = name
        self.args = TexArgs(*args)
        self._contents = contents or []

        for content in contents:
            if isinstance(content, (TexEnv, TexCmd)):
                content.parent = self

    def addContents(self, *contents):
        self._contents.extend(contents)

    @property
    def contents(self):
        """Returns all tokenized chunks for a particular expression."""
        raise NotImplementedError()

    @property
    def tokens(self):
        """Further breaks down all tokens for a particular expression into
        words and other expressions."""
        for content in self.contents:
            if isinstance(str, content):
                for word in content.split():
                    yield word
            yield content

    @property
    def children(self):
        """Returns all child expressions for a particular expression."""
        return filter(lambda x: isinstance(x, (TexEnv, TexCmd)), self.contents)


class TexEnv(TexExpr):
    r"""Abstraction for a LaTeX command, denoted by \begin{env} and \end{env}.
    Contains three attributes: (1) the environment name itself, (2) the
    environment arguments, whether optional or required, and (3) the
    environment's contents.

    >>> t = TexEnv('tabular', ['0 & 0 & * \\\\', '1 & 1 & * \\\\'],
    ...     [RArg('c | c c')])
    >>> t
    TexEnv('tabular', ['0 & 0 & * \\\\', '1 & 1 & * \\\\'], [RArg('c | c c')])
    >>> print(t)
    \begin{tabular}{c | c c}
    0 & 0 & * \\
    1 & 1 & * \\
    \end{tabular}
    >>> len(list(t.children))
    0
    """

    def __init__(self, name, contents=(), args=(), preserve_whitespace=False):
        """Initialization for Tex environment.

        :param str name: name of environment
        :param list contents: list of contents
        :param list args: list of Tex Arguments
        :param bool preserve_whitespace: If false, elements containing only
            whitespace will be removed from contents.
        """
        super().__init__(name, contents, args)

    @property
    def contents(self):
        for content in self._contents:
            if not isinstance(content, str) or bool(content.strip()):
                yield content

    def __str__(self):
        return '\\begin{%s}%s\n%s\n\\end{%s}' % (
            self.name, self.args,'\n'.join(map(str, self._contents)), self.name)

    def __repr__(self):
        if not self.args:
            return "TexEnv('%s')" % self.name
        return "TexEnv('%s', %s, %s)" % (self.name,
            repr(self._contents), repr(self.args))


class TexCmd(TexExpr):
    r"""Abstraction for a LaTeX command. Contains two attributes: (1) the
    command name itself and (2) the command arguments, whether optional or
    required.

    >>> t = TexCmd('textbf',[RArg('big ',TexCmd('textit',[RArg('slant')]),'.')])
    >>> t
    TexCmd('textbf', [RArg('big ', TexCmd('textit', [RArg('slant')]), '.')])
    >>> print(t)
    \textbf{big \textit{slant}.}
    >>> children = list(map(str, t.children))
    >>> len(children)
    1
    >>> print(children[0])
    \textit{slant}
    """

    def __init__(self, name, args=(), extra=''):
        super().__init__(name, [], args)
        self.extra = extra

    @property
    def contents(self):
        """All contents of command arguments"""
        for arg in self.args:
            for expr in arg:
                yield expr
        if self.extra:
            yield self.extra

    def __str__(self):
        if self.extra:
            return '\\%s%s %s' % (self.name, self.args, self.extra)
        return '\\%s%s' % (self.name, self.args)

    def __repr__(self):
        if not self.args:
            return "TexCmd('%s')" % self.name
        return "TexCmd('%s', %s)" % (self.name, repr(self.args))

#############
# Arguments #
#############

class Arg(object):
    """LaTeX command argument"""

    def __init__(self, *exprs):
        """Initialize argument using list of expressions.

        :param [str, TexCmd, TexEnv] exprs: Tex expressions contained in the
            argument. Can be other commands or environments, or even strings.
        """
        self.exprs = exprs

    @property
    def value(self):
        """Argument value, without format."""
        return ''.join(map(str, self.exprs))

    @staticmethod
    def parse(s):
        """Parse a string or list and return an Argument object

        :param (str, list) s: Either a string or a list, where the first and
            last elements are valid argument delimiters.
        """
        if isinstance(s, args):
            return s
        if isinstance(s, (list, tuple)):
            for arg in args:
                if [s[0], s[-1]] == arg.delims():
                    return arg(*s[1:-1])
            raise TypeError('Malformed argument. First and last elements must match a valid argument format: %s' % s)
        for arg in args:
            if arg.__is__(s):
                return arg(arg.__strip__(s))
        raise TypeError('Malformed argument. Must be an Arg or a string in either brackets or curly braces.')

    def __getitem__(self, i):
        """Retrieve an argument's value"""
        return self.value[i]

    @classmethod
    def delims(cls):
        """Returns delimiters"""
        return cls.fmt.split('%s')

    @classmethod
    def __is__(cls, s):
        """Test if string matches format."""
        return s.startswith(cls.delims()[0]) and s.endswith(cls.delims()[1])

    @classmethod
    def __strip__(cls, s):
        """Strip string of format."""
        sides = cls.fmt.split('%s')
        return s[len(cls.delims()[0]):-len(cls.delims()[1])]

    def __iter__(self):
        """Iterator iterates over all argument contents."""
        return iter(self.exprs)

    def __repr__(self):
        """Makes argument display-friendly."""
        return '%s(%s)' % (self.__class__.__name__,
            ', '.join(map(repr, self.exprs)))

    def __str__(self):
        """Stringifies argument value."""
        return self.fmt % self.value

class OArg(Arg):
    """Optional argument."""

    fmt = '[%s]'
    type = 'optional'

class RArg(Arg):
    """Required argument."""

    fmt = '{%s}'
    type = 'required'

args = (OArg, RArg)

class TexArgs(list):
    """List data structure, supporting additional ops for command arguments

    Use regular indexing to access the argument value. Use parentheses, like
    a method invocation, to access an Arg object.

    >>> args = TexArgs(RArg('arg0'), '[arg1]', '{arg2}')
    >>> args
    [RArg('arg0'), OArg('arg1'), RArg('arg2')]
    >>> args(2)
    RArg('arg2')
    >>> args[2]
    'arg2'
    >>> args(2).type
    'required'
    >>> str(args(2))
    '{arg2}'
    >>> args.append('[arg3]')
    >>> args(3)
    OArg('arg3')
    >>> len(args)
    4
    """

    def __init__(self, *args):
        """Append all arguments to list"""
        self.__args = []
        for arg in args:
            self.append(arg)

    def append(self, value):
        """Append a value to the list"""
        arg = Arg.parse(value)
        self.__args.append(arg)
        list.append(self, arg.value)

    def __call__(self, i):
        """
        Access more information about an argument using function-call syntax.
        """
        return self.__args[i]

    def __iter__(self):
        """Iterator iterates over all argument objects."""
        return iter(self.__args)

    def __str__(self):
        """Stringifies a list of arguments.

        >>> str(TexArgs('{a}', '[b]', '{c}'))
        '{a}[b]{c}'
        """
        return ''.join(map(str, self.__args))

    def __repr__(self):
        """Makes list of arguments command-line friendly.

        >>> TexArgs('{a}', '[b]', '{c}')
        [RArg('a'), OArg('b'), RArg('c')]
        """
        return '[%s]' % ', '.join(map(repr, self.__args))



from TexSoup.utils import to_buffer, Buffer
from TexSoup.data import *
import TexSoup.data as data
import functools
import itertools

__all__ = ['read_line', 'read_lines', 'tokenize_line', 'tokenize_lines',
    'read_tex']

WHITESPACE = {' ', '\t', '\r', '\n'}
COMMAND_TOKENS = {'\\'}
ARG_START_TOKENS = {arg.delims()[0] for arg in data.args}
ARG_END_TOKENS = {arg.delims()[1] for arg in data.args}
ARG_TOKENS = ARG_START_TOKENS | ARG_END_TOKENS
ALL_TOKENS = COMMAND_TOKENS | ARG_TOKENS

#######################
# Convenience Methods #
#######################

def read_line(line):
    r"""Read first expression from a single line

    >>> read_line(r'\textbf{Do play \textit{nice}.}')
    TexCmd('textbf', [RArg('Do play ', TexCmd('textit', [RArg('nice')]), '.')])
    >>> print(read_line(r'\newcommand{solution}[1]{{\color{blue} #1}}'))
    \newcommand{solution}[1]{{\color{blue} #1}}
    """
    return read_tex(Buffer(tokenize_line(line)))

def read_lines(*lines):
    r"""Read first expression from multiple lines

    >>> print(read_lines(r'\begin{tabular}{c c}', '0 & 1 \\\\',
    ...     '\end{tabular}'))
    \begin{tabular}{c c}
    0 & 1 \\
    \end{tabular}
    """
    return read_tex(Buffer(itertools.chain(*tokenize_lines(lines))))

#############
# Tokenizer #
#############

@to_buffer
def next_token(line):
    r"""Returns the next possible token, advancing the iterator to the next
    position to start processing from.

    :param (str, iterator, Buffer) line: LaTeX to process
    :return str: the token

    >>> b = Buffer(r'\textbf{Do play\textit{nice}.}')
    >>> print(next_token(b), next_token(b), next_token(b), next_token(b))
    \ textbf { Do play
    >>> print(next_token(b), next_token(b), next_token(b), next_token(b))
    \ textit { nice
    >>> print(next_token(b), next_token(b), next_token(b))
    } . }
    >>> print(next_token(Buffer('.}')))
    .
    >>> next_token(b)
    """
    while line.hasNext():
        for name, f in tokenizers:
            token = f(line)
            if token is not None:
                return token

@to_buffer
def tokenize_line(line):
    r"""Generator for LaTeX tokens on a single line, ignoring comments.

    :param (str, iterator, Buffer) line: LaTeX to process

    >>> print(*tokenize_line(r'\textbf{Do play \textit{nice}.}'))
    \ textbf { Do play  \ textit { nice } . }
    >>> print(*tokenize_line(r'\begin{tabular} 0 & 1 \\ 2 & 0 \end{tabular}'))
    \ begin { tabular }  0 & 1 \\ 2 & 0  \ end { tabular }
    """
    token = next_token(line)
    while token is not None:
        yield token
        token = next_token(line)

def tokenize_lines(lines):
    """Generator for LaTeX tokens across multiple lines, ignoring comments.

    :param list lines: list of strings or iterator over strings
    """
    return map(tokenize_line, lines)

##########
# Tokens #
##########

tokenizers = []

def token(name):
    """Marker for a token

    :param str name: Name of tokenizer
    """
    def wrap(f):
        tokenizers.append((name, f))
        return f
    return wrap

@token('command')
def tokenize_command(line):
    """Process command, but ignore line breaks. (double backslash)

    :param Buffer line: iterator over line, with current position
    """
    if line.peek() == '\\':
        return next(line)

@token('argument')
def tokenize_argument(line):
    """Process both optional and required arguments.

    :param Buffer line: iterator over line, with current position
    """
    for delim in ARG_TOKENS:
        if line.startswith(delim):
            return line.forward(len(delim))

@token('string')
def tokenize_string(line, delimiters=ALL_TOKENS):
    r"""Process a string of text

    :param Buffer line: iterator over line, with current position

    >>> tokenize_string(Buffer('hello'))
    'hello'
    >>> b = Buffer('hello again\command')
    >>> tokenize_string(b)
    'hello again'
    >>> print(b.peek())
    \
    >>> print(tokenize_string(Buffer('0 & 1 \\\\\command')))
    0 & 1 \\
    """
    result = ''
    for c in line:
        if c in delimiters:  # assumes all tokens are single characters
            line.backward(1)
            return result
        result += c
        if line.peek((0, 2)) == '\\\\':
            result += line.forward(2)
    return result

##########
# Mapper #
##########

def read_tex(src):
    r"""Read next expression from buffer

    :param Buffer src: a buffer of tokens
    """
    c = next(src)
    if c == '\\':
        if src.peek().startswith('item '):
            mode, expr = 'command', TexCmd('item', (),
                ' '.join(next(src).split(' ')[1:]))
        elif src.peek() == 'begin':
            mode, expr = next(src), TexEnv(Arg.parse(src.forward(3)).value)
        else:
            mode, expr = 'command', TexCmd(next(src))
        while src.peek() in ARG_START_TOKENS:
            expr.args.append(read_tex(src))
        if mode == 'begin':
            contents = []
            while src.hasNext() and not src.startswith('\\end{%s}' % expr.name):
                if src.peek() in ALL_TOKENS:
                    contents.append(read_tex(src))
                else:
                    contents.append(next(src))
            if not src.startswith('\\end{%s}' % expr.name):
                raise EOFError('Expecting \\end{%s}. Instead got %s' % (
                    expr.name, src.peek((0, 5))))
            else:
                src.forward(5)
            expr.addContents(*contents)
        return expr
    if c in ARG_END_TOKENS:
        return c
    if c in ARG_START_TOKENS:
        content = [c]
        while src.hasNext():
            if src.peek() in ARG_END_TOKENS:
                content.append(next(src))
                break
            elif src.peek() in ALL_TOKENS:
                content.append(read_tex(src))
            else:
                content.append(next(src))
        return Arg.parse(content)



"""
Utils for TexSoup
---

This file consists of various utilities for TexSoup, split into 2 categories by
function:

1. Decorators
2. Generalized utilities
"""

import functools

##############
# Decorators #
##############

def to_buffer(f, i=0):
    """
    Decorator converting all strings and iterators/iterables into
    Buffers.

    :param int i: index of iterator argument. Used only if not a kwarg.
    """
    @functools.wraps(f)
    def wrap(*args, **kwargs):
        iterator = kwargs.get('iterator', args[i])
        if not isinstance(iterator, Buffer):
            iterator = Buffer(iterator)
        if 'iterator' in kwargs:
            kwargs['iterator'] = iterator
        else:
            args = list(args)
            args[i] = iterator
        return f(*args, **kwargs)
    return wrap

#########################
# Generalized Utilities #
#########################

class Buffer:
    """Converts string or iterable into a navigable iterator of strings

    >>> def naturals(i):
    ...   while True:
    ...     yield str(i)
    ...     i += 1
    >>> b1 = Buffer(naturals(0))
    >>> next(b1)
    '0'
    >>> b1.forward()
    '1'
    >>> b1.endswith('1')
    True
    >>> b1.backward(2)
    '01'
    >>> b1.peek()
    '0'
    >>> b1.peek(2)
    '2'
    >>> b1.peek((0, 2))
    '01'
    >>> b1.startswith('01')
    True
    >>> b1[2:4]
    '23'
    >>> b2 = Buffer(naturals(0), coerce=int)
    >>> b2.forward(3)
    12
    >>> Buffer('asdf')[:10]
    'asdf'
    """

    def __init__(self, iterator, coerce=str):
        """Initialization for Buffer

        :param iterator: iterator or iterable
        :param func coerce: optional coercion to datatype (cannot be changed)
        """
        assert hasattr(iterator, '__iter__'), 'Must be an iterable.'
        self.__iterator = iter(iterator)
        self.__queue = []
        self.__i = 0
        self.__coerce = coerce

    def hasNext(self):
        """Returns whether or not there is another element."""
        return bool(self.peek())

    def startswith(self, s):
        """
        Check if iterator starts with s, beginning from the current position
        """
        return self.peek((0, len(s))).startswith(s)

    def endswith(self, s):
        """
        Check if iterator ends with s, ending at current position
        """
        return self.peek((-len(s), 0)).endswith(s)

    def forward(self, j=1):
        """Move forward by j steps."""
        if j < 0:
            return self.backward(-j)
        self.__i += j
        return self[self.__i-j:self.__i]

    def backward(self, j=1):
        """Move backward by j steps."""
        if j < 0:
            return self.forward(-j)
        assert self.__i - j >= 0, 'Cannot move more than %d backward' % self.__i
        self.__i -= j
        return self[self.__i:self.__i+j]

    def peek(self, j=(0, 1)):
        """Peek at the next value(s), without advancing the Buffer"""
        if isinstance(j, int):
            return self[self.__i+j]
        return self[self.__i + j[0]:self.__i + j[1]]

    def __next__(self):
        """Implements next."""
        while self.__i >= len(self.__queue):
            self.__queue.append(self.__coerce(next(self.__iterator)))
        self.__i += 1
        return self.__queue[self.__i-1]

    def __getitem__(self, i):
        """Supports indexing list

        >>> b = Buffer('asdf')
        >>> b[5]
        Traceback (most recent call last):
            ...
        IndexError: list index out of range
        >>> b[0]
        'a'
        >>> b[1:3]
        'sd'
        >>> b[1:]
        'sdf'
        >>> b[:3]
        'asd'
        >>> b[:]
        'asdf'
        """
        j, old = i.stop if not isinstance(i, int) else i, self.__i
        while j is None or self.__i <= j:
            try:
                next(self)
            except StopIteration:
                break
        self.__i = old
        return self.__coerce(''.join(map(str, self.__queue[i])))

    def __iter__(self):
        return self



from TexSoup.reader import *
from TexSoup.data import *
from TexSoup.utils import *
import itertools

def read(tex):
    """Read and parse all LaTeX source

    :param str tex: LaTeX source
    :return TexEnv: the global environment
    """
    if isinstance(tex, str):
        tex = tex.strip().splitlines()
    buf, children = Buffer(itertools.chain(*tokenize_lines(tex))), []
    while buf.hasNext():
        content = read_tex(buf)
        if content is not None:
            children.append(content)
    return TexEnv('[tex]', children)



import sys
from setuptools import setup
from setuptools.command.test import test as TestCommand

class PyTest(TestCommand):

    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)

VERSION = '0.0.1'

setup(
    name = "TexSoup",
    version = VERSION,
    author = "Alvin Wan",
    author_email = 'hi@alvinwan.com',
    description = ("parses valid LaTeX and provides variety of Beautiful-Soup-esque methods and Pythonic idioms for iterating over and searching the parse tree"),
    license = "BSD",
    url = "https://github.com/alvinwan/TexSoup",
    packages = ['TexSoup'],
    cmdclass = {'test': PyTest},
    tests_require = ['pytest'],
    install_requires = [],
    download_url = 'https://github.com/alvinwan/TexSoup/archive/%s.zip' % VERSION,
    classifiers = [
        "Topic :: Utilities",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries",
    ],
)



"""
TexSoup
---

Main file, containing most commonly used elements of TexSoup

@author: Alvin Wan
@site: alvinwan.com
"""

import itertools
import _io
from TexSoup.tex import *

def TexSoup(tex):
    r"""
    At a high-level, parses provided Tex into a navigable, searchable structure.
    This is accomplished in two steps:
    1. Tex is parsed, cleaned, and packaged.
    2. Structure fed to TexNodes for a searchable, coder-friendly interface.

    :param (iterable, string) tex: the Tex source
    :return TexNode: object representing tex document

    >>> from TexSoup import TexSoup
    >>> soup = TexSoup(r'''
    ... \begin{document}
    ...
    ... \section{Hello \textit{world}.}
    ...
    ... \subsection{Watermelon}
    ...
    ... (n.) A sacred fruit. Also known as:
    ...
    ... \begin{itemize}
    ... \item red lemon
    ... \item life
    ... \end{itemize}
    ...
    ... Here is the prevalence of each synonym.
    ...
    ... \begin{tabular}{c c}
    ... red lemon & uncommon \\
    ... life & common
    ... \end{tabular}
    ...
    ... \end{document}
    ... ''')
    >>> soup.section
    \section{Hello \textit{world}.}
    >>> soup.section.name
    'section'
    >>> soup.section.string
    'Hello \\textit{world}.'
    >>> soup.section.parent.name
    'document'
    >>> soup.tabular
    \begin{tabular}{c c}
    red lemon & uncommon \\
    life & common
    \end{tabular}
    >>> soup.tabular.args[0]
    'c c'
    >>> soup.item
    \item red lemon
    >>> list(soup.find_all('item'))
    [\item red lemon, \item life]
    """
    return TexNode(read(tex))
